"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
exports.__esModule = true;
exports.parse = void 0;
var CIRCLE_OBJECT_BIT = 1 << 0;
var SLIDER_OBJECT_BIT = 1 << 1;
var SPINNER_OBJECT_BIT = 1 << 3;
var HOLD_OBJECT_BIT = 1 << 7;
var OBJECT_BITS = CIRCLE_OBJECT_BIT | SLIDER_OBJECT_BIT | SPINNER_OBJECT_BIT | HOLD_OBJECT_BIT;
var DEFAULT_HIT_SAMPLE = '0:0:0:0:';
var parseCurvePoint = function (curvePoint) {
    var _a = curvePoint.split(':'), x = _a[0], y = _a[1];
    return { x: parseInt(x), y: parseInt(y) };
};
var parseHitObject = function (line) {
    var _a = line.split(','), x = _a[0], y = _a[1], time = _a[2], objectType = _a[3], hitSound = _a[4], objectParameters = _a.slice(5);
    var position = { x: parseInt(x), y: parseInt(y) };
    var hitSample = objectParameters.pop() || DEFAULT_HIT_SAMPLE;
    switch (parseInt(objectType) & OBJECT_BITS) {
        case CIRCLE_OBJECT_BIT: {
            return {
                type: 'circle',
                time: parseInt(time),
                position: position,
                hitSample: hitSample
            };
        }
        case SLIDER_OBJECT_BIT: {
            var curve = objectParameters[0], slides = objectParameters[1], _b = objectParameters[2], length_1 = _b === void 0 ? '' : _b, _c = objectParameters[3], edgeSounds = _c === void 0 ? '' : _c, _d = objectParameters[4], edgeSets = _d === void 0 ? '' : _d;
            var _e = curve.split('|'), curveType = _e[0], curvePoints = _e.slice(1);
            return {
                type: 'slider',
                time: parseInt(time),
                position: position,
                hitSample: hitSample,
                curve: {
                    type: curveType,
                    points: curvePoints.map(parseCurvePoint),
                    slides: parseInt(slides),
                    length: parseFloat(length_1),
                    edgeSounds: edgeSounds.split('|').map(function (sound) { return parseInt(sound); }),
                    edgeSets: edgeSets.split('|')
                }
            };
        }
        case SPINNER_OBJECT_BIT: {
            var endTime = objectParameters[0];
            return {
                type: 'spinner',
                time: parseInt(time),
                position: position,
                hitSample: hitSample,
                endTime: parseInt(endTime)
            };
        }
        case HOLD_OBJECT_BIT: {
            throw new Error('HOLD_OBJECT_BIT: unimplemented');
            // return { type: 'hold', time: parseInt(time), position }
        }
        default:
            throw new Error('Invalid hit object type');
    }
};
var parseTimingPoint = function (line) {
    var _a = line.split(','), time = _a[0], beatLength = _a[1], meter = _a[2], sampleSet = _a[3], sampleIndex = _a[4], volume = _a[5], uninherited = _a[6], effects = _a[7];
    return {
        time: parseInt(time),
        beatLength: parseFloat(beatLength),
        meter: parseInt(meter),
        sampleSet: parseInt(sampleSet),
        sampleIndex: parseInt(sampleIndex),
        volume: parseInt(volume),
        uninherited: uninherited === '1',
        effects: parseInt(effects)
    };
};
var parseKeyValuePair = function (line) {
    var _a = line.split(':'), key = _a[0], _b = _a[1], value = _b === void 0 ? '' : _b;
    return [key.trim(), value.trim()];
};
var GENERAL_KEY_MAP = {
    AudioFilename: { key: 'audioFilename', type: 'string' },
    AudioLeadIn: { key: 'audioLeadIn', type: 'integer', "default": 0 },
    PreviewTime: { key: 'previewTime', type: 'integer', "default": -1 },
    Countdown: { key: 'countdown', type: 'integer', "default": 1 },
    SampleSet: { key: 'sampleSet', type: 'string', "default": 'Normal' },
    StackLeniency: { key: 'stackLeniency', type: 'decimal', "default": 0.7 },
    Mode: { key: 'mode', type: 'integer', "default": 0 },
    LetterboxInBreaks: {
        key: 'letterboxInBreaks',
        type: 'boolean',
        "default": false
    },
    UseSkinSprites: { key: 'useSkinSprites', type: 'boolean', "default": false },
    OverlayPosition: {
        key: 'overlayPosition',
        type: 'string',
        "default": 'NoChange'
    },
    SkinPreference: { key: 'skinPreference', type: 'string', "default": '' },
    EpilepsyWarning: { key: 'epilepsyWarning', type: 'boolean', "default": false },
    CountdownOffset: { key: 'countdownOffset', type: 'integer', "default": 0 },
    SpecialStyle: { key: 'specialStyle', type: 'boolean', "default": false },
    WidescreenStoryboard: {
        key: 'widescreenStoryboard',
        type: 'boolean',
        "default": false
    },
    SamplesMatchPlaybackRate: {
        key: 'samplesMatchPlaybackRate',
        type: 'boolean',
        "default": false
    }
};
var DIFFICULTY_KEY_MAP = {
    HPDrainRate: { key: 'hp', type: 'decimal', "default": -1 },
    CircleSize: { key: 'cs', type: 'decimal', "default": -1 },
    OverallDifficulty: { key: 'od', type: 'decimal', "default": -1 },
    ApproachRate: { key: 'ar', type: 'decimal', "default": -1 },
    SliderMultiplier: { key: 'sliderMultiplier', type: 'decimal', "default": -1 },
    SliderTickRate: { key: 'sliderTickRate', type: 'decimal', "default": -1 }
};
var METADATA_KEY_MAP = {
    Title: { key: 'title', type: 'string' },
    TitleUnicode: { key: 'titleUnicode', type: 'string' },
    Artist: { key: 'artist', type: 'string' },
    ArtistUnicode: { key: 'artistUnicode', type: 'string' },
    Creator: { key: 'creator', type: 'string' },
    Version: { key: 'version', type: 'string' },
    Source: { key: 'source', type: 'string' },
    Tags: { key: 'tags', type: 'string' },
    BeatmapID: { key: 'id', type: 'integer' },
    BeatmapSetID: { key: 'setId', type: 'integer' }
};
var parseKeyValuePairSection = function (lines, sectionKeyMap) {
    var section = {};
    for (var key in sectionKeyMap) {
        var defaultValue = sectionKeyMap[key]["default"];
        if (defaultValue !== undefined)
            section[sectionKeyMap[key].key] = defaultValue;
    }
    for (var _i = 0, lines_1 = lines; _i < lines_1.length; _i++) {
        var line = lines_1[_i];
        var _a = parseKeyValuePair(line), key = _a[0], value = _a[1];
        var sectionKey = sectionKeyMap[key];
        if (!sectionKey)
            throw new Error("Unhandled key " + key);
        switch (sectionKey.type) {
            case 'integer':
                section[sectionKey.key] = parseInt(value);
                break;
            case 'decimal':
                section[sectionKey.key] = parseFloat(value);
                break;
            case 'string':
                section[sectionKey.key] = value;
                break;
            case 'boolean':
                section[sectionKey.key] = value === '1';
                break;
        }
    }
    return section;
};
var GAME_MODES = ['osu', 'taiko', 'catch', 'mania'];
var parse = function (data) {
    var lines = data
        .split('\n')
        .filter(function (line) {
        var isComment = line.startsWith(' ') || line.startsWith('_') || line.startsWith('//');
        var isEmpty = line.trim() === '';
        return !isComment && !isEmpty;
    })
        .map(function (line) { return line.trim(); });
    var header = lines.splice(0, 1)[0];
    var version = parseInt(header.replace(/[\d]+/g, ''));
    var sectionLines = {
        General: [],
        Editor: [],
        Metadata: [],
        Difficulty: [],
        Events: [],
        TimingPoints: [],
        Colours: [],
        HitObjects: []
    };
    var section = null;
    for (var _i = 0, lines_2 = lines; _i < lines_2.length; _i++) {
        var line = lines_2[_i];
        if (line.startsWith('[')) {
            section = line.substr(1, line.length - 2);
            continue;
        }
        if (!section)
            continue;
        sectionLines[section].push(line);
    }
    var beatmapBase = {
        version: version,
        general: parseKeyValuePairSection(sectionLines.General, GENERAL_KEY_MAP),
        metadata: parseKeyValuePairSection(sectionLines.Metadata, METADATA_KEY_MAP),
        difficulty: parseKeyValuePairSection(sectionLines.Difficulty, DIFFICULTY_KEY_MAP),
        timingPoints: sectionLines.TimingPoints.map(function (line) {
            return parseTimingPoint(line);
        }),
        hitObjects: sectionLines.HitObjects.map(function (line) { return parseHitObject(line); })
    };
    // Support old beatmaps where AR is not set
    if (beatmapBase.difficulty.ar < 0) {
        beatmapBase.difficulty.ar = beatmapBase.difficulty.od;
    }
    return __assign({ mode: GAME_MODES[beatmapBase.general.mode], id: beatmapBase.metadata.id, setId: beatmapBase.metadata.setId }, beatmapBase);
};
exports.parse = parse;
