import { Vec2 } from './vec2';
export declare type General = Readonly<{
    audioFilename: string;
    audioLeadIn: number;
    previewTime: number;
    countdown: number;
    sampleSet: string;
    stackLeniency: number;
    mode: number;
    letterboxInBreaks: boolean;
    useSkinSprites: boolean;
    overlayPosition: string;
    skinPreference: string;
    epilepsyWarning: boolean;
    countdownOffset: number;
    specialStyle: boolean;
    widescreenStoryboard: boolean;
    samplesMatchPlaybackRate: boolean;
}>;
export declare type Metadata = Readonly<{
    title: string;
    titleUnicode: string;
    artist: string;
    artistUnicode: string;
    creator: string;
    version: string;
    source: string;
    tags: string;
    id: number;
    setId: number;
}>;
export declare type Difficulty = {
    hp: number;
    cs: number;
    od: number;
    ar: number;
    sliderMultiplier: number;
    sliderTickRate: number;
};
export declare type BeatmapBase = Readonly<{
    version: number;
    general: General;
    metadata: Metadata;
    difficulty: Difficulty;
    timingPoints: TimingPoint[];
    hitObjects: HitObject[];
}>;
export declare type OsuBeatmap = Readonly<{
    mode: 'osu';
    id: number;
    setId: number;
}> & BeatmapBase;
export declare type TaikoBeatmap = Readonly<{
    mode: 'taiko';
    id: number;
    setId: number;
}> & BeatmapBase;
export declare type ManiaBeatmap = Readonly<{
    mode: 'mania';
    id: number;
    setId: number;
}> & BeatmapBase;
export declare type CatchBeatmap = Readonly<{
    mode: 'catch';
    id: number;
    setId: number;
}> & BeatmapBase;
export declare type Beatmap = OsuBeatmap | TaikoBeatmap | ManiaBeatmap | CatchBeatmap;
export declare type TimingPoint = Readonly<{
    time: number;
    beatLength: number;
    meter: number;
    sampleSet: number;
    sampleIndex: number;
    volume: number;
    uninherited: boolean;
    effects: number;
}>;
export declare type CircleObject = Readonly<{
    type: 'circle';
    position: Vec2;
    time: number;
    hitSample: string;
}>;
export declare type SliderCurve = Readonly<{
    type: 'B' | 'C' | 'L' | 'P';
    points: Vec2[];
    slides: number;
    length: number;
    edgeSounds: number[];
    edgeSets: string[];
}>;
export declare type SliderObject = Readonly<{
    type: 'slider';
    position: Vec2;
    time: number;
    curve: SliderCurve;
    hitSample: string;
}>;
export declare type SpinnerObject = Readonly<{
    type: 'spinner';
    position: Vec2;
    hitSample: string;
    time: number;
    endTime: number;
}>;
export declare type HoldObject = Readonly<{
    type: 'hold';
    position: Vec2;
    time: number;
}>;
export declare type HitObject = CircleObject | SliderObject | SpinnerObject | HoldObject;
